import React, { useEffect } from 'react';
import { createEditor, Transforms, Editor, Node } from 'slate';
import { withHistory } from 'slate-history';
import { withReact, ReactEditor, DefaultElement, Slate, Editable } from 'slate-react';
import { ulid } from 'ulid';

function _extends() {
  _extends = Object.assign ? Object.assign.bind() : function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];
      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }
    return target;
  };
  return _extends.apply(this, arguments);
}

function debounce(func, waitMilliseconds = 50, options = {}) {
  var _options$isImmediate, _options$callback;
  let timeoutId;
  const isImmediate = (_options$isImmediate = options.isImmediate) != null ? _options$isImmediate : false;
  const callback = (_options$callback = options.callback) != null ? _options$callback : false;
  const maxWait = options.maxWait;
  let lastInvokeTime = Date.now();
  let promises = [];
  function nextInvokeTimeout() {
    if (maxWait !== undefined) {
      const timeSinceLastInvocation = Date.now() - lastInvokeTime;
      if (timeSinceLastInvocation + waitMilliseconds >= maxWait) {
        return maxWait - timeSinceLastInvocation;
      }
    }
    return waitMilliseconds;
  }
  const debouncedFunction = function debouncedFunction(...args) {
    const context = this;
    return new Promise((resolve, reject) => {
      const invokeFunction = function invokeFunction() {
        timeoutId = undefined;
        lastInvokeTime = Date.now();
        if (!isImmediate) {
          const result = func.apply(context, args);
          callback && callback(result);
          promises.forEach(({
            resolve
          }) => resolve(result));
          promises = [];
        }
      };
      const shouldCallNow = isImmediate && timeoutId === undefined;
      if (timeoutId !== undefined) {
        clearTimeout(timeoutId);
      }
      timeoutId = setTimeout(invokeFunction, nextInvokeTimeout());
      if (shouldCallNow) {
        const result = func.apply(context, args);
        callback && callback(result);
        return resolve(result);
      }
      promises.push({
        resolve,
        reject
      });
    });
  };
  debouncedFunction.cancel = function (reason) {
    if (timeoutId !== undefined) {
      clearTimeout(timeoutId);
    }
    promises.forEach(({
      reject
    }) => reject(reason));
    promises = [];
  };
  return debouncedFunction;
}

/**
 * Returns a random integer between min (inclusive) and max (inclusive).
 * The value is no lower than min (or the next integer greater than min
 * if min isn't an integer) and no greater than max (or the next integer
 * lower than max if max isn't an integer).
 * Using Math.round() will give you a non-uniform distribution!
 */
function getRandomInt(min, max) {
  min = Math.ceil(min);
  max = Math.floor(max);
  return Math.floor(Math.random() * (max - min + 1)) + min;
}

function throttle(func, limit) {
  let inThrottle;
  let lastResult;
  return function () {
    const args = arguments;
    const context = this;
    if (!inThrottle) {
      inThrottle = true;
      setTimeout(() => inThrottle = false, limit);
      lastResult = func.apply(context, args);
    }
    return lastResult;
  };
}

/**
 * Responsible for fetching completions from a model and adding them to the input.
 * Each `<SmartInput />` component rendered to the DOM will have its own instance of this class as state.
 */
class SmartInputFetcher {
  /**
   * sequenceId is used to ensure that we don't add completions from a previous fetch,
   * we increment it each time we request to fetch a new completion because the user input has changed
   */

  /**
   * sessionId is used to track this particular instantiation of the `<SmartInput />` component
   */

  /**
   * completionCache is used to cache the completions that we've already fetched from the model
   */

  /**
   * alreadySeen tracks which completions we've marked as seen (reduce chattiness)
   */

  /**
   * alreadyAccepted tracks which completions we've marked as acceoted (reduce chattiness)
   */

  /**
   * currentCompletion tracks the latest text that was sent to the model for fetching completions, and the
   * completion that was returned (if any ~ we may still be waiting for it)
   */

  constructor(apiUrl, addCompletion, modelId, mockCompletion) {
    var _this = this;
    this.apiUrl = void 0;
    this.addCompletion = void 0;
    this.modelId = void 0;
    this.mockCompletion = void 0;
    this.sequenceId = 0;
    this.sessionId = ulid();
    this.completionCache = {};
    this.alreadySeen = new Set();
    this.alreadyAccepted = new Set();
    this.currentCompletion = null;
    this.fetchCompletion = debounce(async function (text, context) {
      var _this$currentCompleti, _this$currentCompleti3;
      if (text.length === 0) {
        // only get completions for non-empty text
        return null;
      }
      const seqId = ++_this.sequenceId;
      // if a completion was already added, and the user is typing the same characters as the current completion, don't fetch
      if ((_this$currentCompleti = _this.currentCompletion) != null && _this$currentCompleti.completion) {
        const currExpectedText = _this.currentCompletion.text + _this.currentCompletion.completion;
        if (currExpectedText.startsWith(text)) {
          return null;
        }
      }
      // update the current completion
      _this.currentCompletion = {
        text,
        sequenceId: seqId
      };
      // if we already have a completion for this text, don't fetch, but do return it
      const cachedVal = _this.completionCache[text];
      if (cachedVal) {
        // don't fetch the same text twice
        _this.addCompletion(text, cachedVal);
        _this.currentCompletion.completion = cachedVal.completion;
        return cachedVal;
      }
      if (_this.mockCompletion) {
        const completion = _this.mockCompletion(text);
        if (completion) {
          return await new Promise(resolve => {
            setTimeout(() => {
              var _this$currentCompleti2;
              const c = {
                input: text,
                sid: ulid(),
                completion
              };
              _this.completionCache[text] = c;
              if (((_this$currentCompleti2 = _this.currentCompletion) == null ? void 0 : _this$currentCompleti2.sequenceId) === seqId) {
                _this.addCompletion(text, c);
                _this.currentCompletion.completion = completion;
              }
              resolve(c);
            }, getRandomInt(200, 1000)); // simulate API latency
          });
        }

        return {
          input: text,
          sid: ulid(),
          completion: null
        };
      }
      const res = await fetch(`${_this.apiUrl}/completions`, {
        method: 'POST',
        mode: 'cors',
        headers: {
          'Content-Type': 'application/json',
          Authorization: `Bearer ${_this.modelId}`
        },
        body: JSON.stringify({
          model: _this.modelId,
          input: text,
          session: _this.sessionId,
          context
        })
      });
      if (res.status !== 200) {
        console.error("Couldn't fetch completions");
        return null;
      }
      const data = await res.json();
      _this.completionCache[text] = data;
      if (((_this$currentCompleti3 = _this.currentCompletion) == null ? void 0 : _this$currentCompleti3.sequenceId) === seqId) {
        _this.addCompletion(text, data);
        _this.currentCompletion.completion = data.completion;
      }
      return data;
    }, 200);
    this.sendEvents = throttle((text, context) => {
      fetch(`${this.apiUrl}/events`, {
        method: 'POST',
        mode: 'cors',
        headers: {
          'Content-Type': 'application/json',
          Authorization: `Bearer ${this.modelId}`
        },
        body: JSON.stringify({
          model: this.modelId,
          session: this.sessionId,
          input: text,
          context
        })
      }).then(res => {
        if (res.status !== 200) {
          console.error('Error sending events');
        }
      }).catch(e => console.error(e));
    }, 1000);
    this.apiUrl = apiUrl;
    this.addCompletion = addCompletion;
    this.modelId = modelId;
    this.mockCompletion = mockCompletion;
  }
  updateCompletion(completionSid, arg) {
    if (this.mockCompletion) {
      return;
    }
    // reduce chattiness
    const shouldUpdate = arg.seen && !this.alreadySeen.has(completionSid) || arg.accepted && !this.alreadyAccepted.has(completionSid);
    if (!shouldUpdate) {
      return;
    }
    fetch(`${this.apiUrl}/completions`, {
      method: 'PUT',
      mode: 'cors',
      headers: {
        'Content-Type': 'application/json',
        Authorization: `Bearer ${this.modelId}`
      },
      body: JSON.stringify({
        completionSid,
        seen: arg.seen,
        accepted: arg.accepted
      })
    }).then(res => {
      if (res.status !== 200) {
        console.error("Couldn't update completion");
      }
      if (arg.seen) {
        // only update the set if the update was successful
        this.alreadySeen.add(completionSid);
      }
      if (arg.accepted) {
        // only update the set if the update was successful
        this.alreadyAccepted.add(completionSid);
      }
    }).catch(e => console.error(e));
  }
}

console.log('here yep yep');
/**
 * A text input that provides intelligent, tab-completable completions based on the text that has already been entered.
 * Due to the limitations of `<input>` and `<textarea>` elements, this component renders a `<div contenteditable="true">` instead.
 */
function SmartInput(props) {
  const editor = React.useMemo(() => withHistory(withReact(createEditor())), []);
  const [completion, setCompletion] = React.useState(null);
  const [fetcher] = React.useState(new SmartInputFetcher(props.devApiUrl || 'https://api.foundation-ui.com/smartinput', (text, completion) => {
    addCompletion(completion);
  }, props.modelId, props.modelId ? props.mockCompletion : () => ' -- sample completion --'));
  const acceptCompletion = () => {
    EditorHelper.acceptCompletion(editor, fetcher, completion);
    setCompletion(null);
  };
  const addCompletion = completion => {
    if (props.value) {
      // don't add completion if the text is gone
      EditorHelper.addCompletion(editor, fetcher, completion);
      setCompletion(completion);
    }
  };
  const clearCompletion = () => {
    EditorHelper.removeCompletion(editor);
    setCompletion(null);
  };
  // for good measure, always overwrite the fetcher fields
  // fetcher.modelId = props.modelId
  fetcher.addCompletion = (text, completion) => {
    const currText = props.value;
    if (completion && (text + completion).startsWith(currText)) {
      addCompletion(completion);
    }
  };
  // fetcher.mockCompletion = props.mockCompletion
  const initialValue = [{
    type: 'text',
    children: [{
      text: props.value,
      type: undefined
    }]
  }];
  useEffect(() => {
    if (props.editorRef) {
      props.editorRef.current = _extends({}, ReactEditor.toDOMNode(editor, editor), {
        focus: () => {
          setTimeout(() => {
            ReactEditor.focus(editor);
            if (props.value) {
              Transforms.select(editor, Editor.end(editor, []));
            } else {
              Transforms.select(editor, Editor.start(editor, []));
            }
          }, 50);
        }
      });
    }
  });
  const renderElement = React.useCallback(p => {
    switch (p.element.type) {
      case 'text':
        return props.renderText(_extends({}, p));
      case 'completion':
        return props.renderCompletion(_extends({}, p, {
          contentEditable: 'false',
          suppressContentEditableWarning: true
        }));
      default:
        return /*#__PURE__*/React.createElement(DefaultElement, p);
    }
  },
  // eslint-disable-next-line react-hooks/exhaustive-deps
  [props.renderText, props.renderCompletion]);
  const renderPlaceholder = React.useCallback(p => {
    return props.renderPlaceholder(_extends({}, p, p.attributes || {}, {
      style: _extends({
        // these style attributes usually come from slate, we gotta make sure to add them back or the cursor might go after the placeholder lol
        position: 'absolute',
        pointerEvents: 'none',
        width: '100%',
        maxWidth: '100%',
        userSelect: 'none',
        textDecoration: 'none'
      }, p.style || {})
    }));
  }, [props.renderPlaceholder]);
  return /*#__PURE__*/React.createElement(Slate, {
    editor: editor,
    value: initialValue,
    onChange: value => {
      var _fetcher$fetchComplet;
      const text = EditorHelper.serialize(value.filter(n => n.type !== 'completion'));
      (_fetcher$fetchComplet = fetcher.fetchCompletion(text, props.context || {})) == null ? void 0 : _fetcher$fetchComplet.catch(e => console.error(e));
      fetcher.sendEvents(text, props.context || {});
      props.onChange(text);
    }
  }, /*#__PURE__*/React.createElement(Editable, {
    placeholder: props.placeholder,
    className: props.className,
    style: _extends({}, props.style || {}, {
      whiteSpace: props.multiline ? undefined : 'pre'
    }),
    onPaste: event => {
      clearCompletion();
      event.preventDefault();
      const value = event.clipboardData.getData('Text');
      if (!props.multiline) {
        editor.insertText(value.replaceAll('\n', ''));
      } else {
        editor.insertText(value);
      }
    },
    renderPlaceholder: props.renderPlaceholder ? renderPlaceholder : undefined,
    renderElement: renderElement,
    onBlur: e => {
      clearCompletion();
      fetcher.currentCompletion = null;
      if (props.onBlur) {
        props.onBlur(e);
      }
    },
    onFocus: e => {
      if (props.onFocus) {
        props.onFocus(e);
        if (editor.selection == null) {
          // fix bug for chrome
          if (props.value) {
            Transforms.select(editor, Editor.end(editor, []));
          } else {
            Transforms.select(editor, Editor.start(editor, []));
          }
        }
      }
    },
    onKeyDown: event => {
      var _completion$completio, _completion$completio2;
      const hasCompletion = Boolean(completion == null ? void 0 : (_completion$completio = completion.completion) == null ? void 0 : (_completion$completio2 = _completion$completio.trim()) == null ? void 0 : _completion$completio2.length);
      if (event.key === 'Tab' && hasCompletion) {
        event.preventDefault();
        acceptCompletion();
        fetcher.currentCompletion = null;
      } else if (event.key === 'ArrowRight' && hasCompletion) {
        event.preventDefault();
        acceptCompletion();
        fetcher.currentCompletion = null;
      } else if (event.key === 'ArrowLeft' && hasCompletion) {
        clearCompletion();
        fetcher.currentCompletion = null;
      } else if ((event.key === 'ArrowDown' || event.key === 'ArrowUp') && hasCompletion) {
        clearCompletion();
      } else if (event.key === 'Backspace') {
        fetcher.currentCompletion = null;
        clearCompletion();
      } else if (event.key === 'Enter' && !event.metaKey) {
        event.preventDefault();
        if (props.multiline) {
          clearCompletion();
          editor.insertText('\n');
        }
      } else if (event.key !== 'Shift') {
        var _completion$completio3;
        if (completion && !event.metaKey && (_completion$completio3 = completion.completion) != null && _completion$completio3.startsWith(event.key) && event.key.length === 1) {
          const newCompletionText = completion.completion.substring(1);
          // don't clear the completion if the user is just typing the completion character for character
          EditorHelper.updateCompletion(editor, newCompletionText);
          setCompletion(_extends({}, completion, {
            completion: newCompletionText
          }));
        } else {
          fetcher.currentCompletion = null;
          clearCompletion(); // TODO: this is a bit aggressive?
        }
      }

      if (props.onKeyDown) {
        props.onKeyDown(event);
      }
    }
  }));
}
const EditorHelper = {
  isCompletionActive: editor => {
    const completion = editor.children.find(n => n.type === 'completion');
    return !!completion;
  },
  isCursorAtEndOfDocument: editor => {
    const sel = editor.selection;
    if (!sel) {
      return false;
    }
    const [node] = sel.anchor.path;
    const offset = sel.anchor.offset;
    if (node !== editor.children.length - 1) {
      // not at the last node
      return false;
    }
    return offset === Node.string(editor.children[node]).length;
  },
  acceptCompletion: (editor, fetcher, completion) => {
    const isActive = EditorHelper.isCompletionActive(editor);
    if (isActive && completion != null && completion.completion) {
      EditorHelper.removeCompletion(editor);
      editor.insertText(completion.completion);
      fetcher.updateCompletion(completion.sid, {
        accepted: true
      });
    }
  },
  addCompletion(editor, fetcher, completion) {
    const isActive = EditorHelper.isCompletionActive(editor);
    const isCursorEnd = EditorHelper.isCursorAtEndOfDocument(editor);
    if (!isActive && isCursorEnd && completion != null && completion.completion) {
      Transforms.insertNodes(editor, {
        type: 'completion',
        children: [{
          text: completion.completion
        }]
      }, {
        at: [editor.children.length]
      });
      fetcher.updateCompletion(completion.sid, {
        seen: true
      });
    }
  },
  removeCompletion(editor) {
    const isActive = EditorHelper.isCompletionActive(editor);
    if (isActive) {
      Transforms.removeNodes(editor, {
        at: [editor.children.length - 1]
      });
    }
  },
  updateCompletion(editor, newCompletion) {
    const isActive = EditorHelper.isCompletionActive(editor);
    if (isActive && newCompletion.length > 0) {
      // could not find a way to update to set the text of the node, so we remove the first character only
      Transforms.delete(editor, {
        at: {
          anchor: {
            path: [editor.children.length - 1, 0],
            offset: 0
          },
          focus: {
            path: [editor.children.length - 1, 0],
            offset: 1
          }
        }
      });
    }
  },
  serialize(nodes) {
    return nodes.map(n => Node.string(n)).join('\n');
  }
};

// export * from './SmartInput'

export { SmartInput as default };
//# sourceMappingURL=index.modern.js.map
